"""
This file provides information of how to build and configure CERN.ROOT:
https://github.com/root-project/root

"""

import os
from distutils.dir_util import mkpath

from ejpm.engine import env_gen
from ejpm.engine.db import BUILT_WITH_CONFIG
from ejpm.engine.recipe import Recipe
from ejpm.engine.commands import run, env, workdir, is_not_empty_dir

ROOTSYS = "ROOTSYS"


class RootRecipe(Recipe):
    """Provides data for building and installing root

    PackageInstallationContext is located in engine/recipe.py

    """

    class DefaultConfigFields(object):
        pass

    def __init__(self):
        """
        """

        # Fill the common path pattern
        super(RootRecipe, self).__init__("root")
        self.config['branch'] = 'v6-20-04'
        self.config['cmake_custom_flags'] = ''
        self.config['cmake_build_type']='RelWithDebInfo'

    def find_python(self):
        """Searches default python which is first found in PATH"""
        from subprocess import check_output
        out = check_output(["which", "python3"]).decode('ascii').strip()

        if not out:
            out = check_output(["which", "python2"]).decode('ascii').strip()

        if not out:
            out = check_output(["which", "python"]).decode('ascii').strip()
        return out

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        #
        # Compile with python3, then whatever python is...
        python_path = self.find_python()
        self.config["python_flag"] = ' -DPYTHON_EXECUTABLE={} '.format(python_path) if python_path else ''
        # >oO debug: print("Configuring ROOT with '{}' python flag".format(self.config["python_flag"]))

        #
        # use_common_dirs_scheme sets standard package variables:
        # version      = 'v{}-{:02}-{:02}'                 # Stringified version. Used to create directories and so on
        # source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{version}         # Where the binary installation is
        self.use_common_dirs_scheme()

        #
        # Root download link. We will use github root mirror:
        # The tags have names like: v6-14-04
        # http://github.com/root-project/root.git
        # clone with shallow copy
        self.clone_command = "git clone --depth 1 -b {branch} https://github.com/root-project/root.git {source_path}" \
            .format(**self.config)

        # Make sure custom flags are in there
        if "cmake_custom_flags" not in self.config:
            self.config["cmake_custom_flags"] = ''

        #
        # ROOT packets to disable in our build (go with -D{name}=ON flag)
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev -DCMAKE_INSTALL_PREFIX={install_path} " \
                         " -DCMAKE_CXX_STANDARD={cxx_standard}" \
                         " -DCMAKE_BUILD_TYPE={cmake_build_type}"\
                         " -Dhttp=ON" \
                         " -Droot7=ON" \
                         " -Dgdml=ON" \
                         " -Dminuit2=ON" \
                         " {python_flag}" \
                         " {cmake_custom_flags}" \
                         " {source_path}" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
            .format(**self.config)  # make global options like '-j8'. Skip now

    def step_install(self):
        self.step_clone_root()
        self.step_build_root()

    def step_clone_root(self):
        """Clones root from github mirror"""

        # Check the directory exists and not empty
        if is_not_empty_dir(self.source_path):
            return  # The directory exists and is not empty. Assume it cloned

        mkpath(self.source_path)  # Create the directory and any missing ancestor directories if not there
        run(self.clone_command)  # Execute git clone command

    def step_build_root(self):
        """Builds root from the ground"""

        # Create build directory
        run('mkdir -p {}'.format(self.build_path))

        env('ROOTSYS', self.install_path)

        # go to our build directory
        workdir(self.build_path)

        # run cmake && make && install
        run(self.build_cmd)

    def step_rebuild_root(self):
        """Clear root build directory"""

        # clear sources directories if needed
        run('rm -rf {}'.format(self.source_path))
        run('rm -rf {}'.format(self.build_path))

        # Now run build root
        self.step_build_root()

    @staticmethod
    def gen_env(data):
        install_path = data['install_path']

        isinstance(data, dict)

        # The next is about conda
        # in conda thisroot.sh triggers error explaining, that everything is already done in activate
        # so we don't need to put thisroot if we acting under conda
        # this is hacky hack
        is_under_conda = 'ROOT_INSTALLED_BY_CONDA' in os.environ

        # In any case we need python environment to build stuff with root under ejpm
        def update_python_environment():
            """Function that will update ROOT environment in python
            We need this function because we DON'T want source thisroot in python
            """

            root_bin = os.path.join(install_path, 'bin')
            root_lib = os.path.join(install_path, 'lib')
            root_jup = os.path.join(install_path, 'etc', 'notebook')

            env_updates = [
                env_gen.Set('ROOTSYS', install_path),
                env_gen.Prepend('PATH', root_bin),
                env_gen.Prepend('LD_LIBRARY_PATH', root_lib),
                env_gen.Prepend('DYLD_LIBRARY_PATH', root_lib),
                env_gen.Prepend('PYTHONPATH', root_lib),
                env_gen.Prepend('CMAKE_PREFIX_PATH', install_path),
                env_gen.Prepend('JUPYTER_PATH', root_jup),
            ]

            for updater in env_updates:
                updater.update_python_env()

        # We just call thisroot.xx in different shells
        bash_thisroot_path = os.path.join(install_path, 'bin', 'thisroot.sh')
        bash_text =  'if test -f "{0}"; then source {0}; fi'.format(bash_thisroot_path)

        csh_thisroot_path = os.path.join(install_path, 'bin', 'thisroot.csh')
        csh_text = "if ( -f {0} ) then\n"\
                   "    source {0}\n"\
                   "endif"\
                   .format(csh_thisroot_path)

        bash_text = bash_text if not is_under_conda else "# Don't call thisroot.sh under conda"
        csh_text = csh_text if not is_under_conda else "# Don't call thisroot.csh under conda"
        # Do we need this because of conda?

        # RawText requires text for bash, csh and a function for python
        raw = env_gen.RawText(bash_text, csh_text, update_python_environment)
        yield raw

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    #                   }
    os_dependencies = {
        'required': {
            'ubuntu': "dpkg-dev binutils libx11-dev libxpm-dev libxft-dev libxext-dev",
            'centos': "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel",
            'centos8': "gcc binutils libX11-devel libXpm-devel libXft-devel libXext-devel"
        },
        'optional': {
            'ubuntu': "gfortran libssl-dev libpcre3-dev "
                      "xlibmesa-glu-dev libglew1.5-dev libftgl-dev "
                      "libmysqlclient-dev libfftw3-dev libcfitsio-dev "
                      "graphviz-dev libavahi-compat-libdnssd-dev "
                      "libldap2-dev python-dev libxml2-dev libkrb5-dev "
                      "libgsl0-dev libqt4-dev",

            'centos': "gcc-gfortran openssl-devel pcre-devel "
                      "mesa-libGL-devel mesa-libGLU-devel glew-devel ftgl-devel mysql-devel "
                      "fftw-devel cfitsio-devel graphviz-devel "
                      "avahi-compat-libdns_sd-devel libldap-dev python-devel "
                      "libxml2-devel gsl-static",
            'centos8': "gcc-gfortran openssl-devel pcre-devel "
                      "mesa-libGL-devel mesa-libGLU-devel glew-devel ftgl-devel mysql-devel "
                      "fftw-devel cfitsio-devel graphviz-devel "
                      "avahi-compat-libdns_sd-devel openldap-devel python36-devel "
                      "libxml2-devel gsl-devel"
        },
    }

    # (!) Those are not used at the moment. But there is an idea to have minimized version of root
    root_mini_flags = [
        "afdsmgrd           = OFF",  # Dataset manager for PROOF-based analysis facilities
        "afs                = OFF",  # AFS support, requires AFS libs and objects
        "alien              = OFF",  # ON  - AliEn support, requires libgapiUI from ALICE
        "all                = OFF",  # OFF - Enable all optional components
        "asimage            = OFF",  # ON  - Image processing support, requires libAfterImage
        "astiff             = OFF",  # ON  - Include tiff support in image processing
        "bonjour            = OFF",  # ON  - Bonjour support, requires libdns_sd and/or Avahi
        "builtin_afterimage = ON ",  # ON  - Built included libAfterImage, or use system libAfterImage
        "builtin_fftw3      = ON",  # OFF - Built the FFTW3 library internally (downloading tarfile from the Web)
        "builtin_ftgl       = ON ",  # ON  - Built included libFTGL, or use system libftgl
        "builtin_freetype   = OFF",  # OFF - Built included libfreetype, or use system libfreetype
        "builtin_glew       = ON ",  # ON  - Built included libGLEW, or use system libGLEW
        "builtin_pcre       = OFF",  # OFF - Built included libpcre, or use system libpcre
        "builtin_zlib       = OFF",  # OFF - Built included libz, or use system libz
        "builtin_lzma       = OFF",  # OFF - Built included liblzma, or use system liblzma
        "builtin_davix      = OFF",  # OFF - Built the Davix library internally (downloading tarfile from the Web)
        "builtin_gsl        = OFF",  # OFF - Built the GSL library internally (downloading tarfile from the Web)
        "builtin_cfitsio    = OFF",  # OFF - Built the FITSIO library internally (downloading tarfile from the Web)
        "builtin_xrootd     = OFF",  # OFF - Built the XROOTD internally (downloading tarfile from the Web)
        "builtin_llvm       = ON ",  # ON  - Built the LLVM internally
        "builtin_tbb        = ON ",  # OFF - Built the TBB internally

        "cxx11              = ON ",  # ON  - Build using C++11 compatible mode, requires gcc > 4.7.x or clang
        "cxx14              = OFF",  # OFF - Build using C++14 compatible mode, requires gcc > 4.9.x or clang
        "cxx17              = OFF",  # OFF - Build using C++17 compatible mode, requires gcc > 7.1.x or clang

        "libcxx             = OFF",  # OFF - Build using libc++, requires cxx11 option
        "castor             = OFF",  # ON  - CASTOR support, requires libshift from CASTOR >               =  1.5.2
        "ccache             = OFF",  # OFF - Enable ccache usage for speeding up builds
        "chirp              = OFF",  # ON  - Chirp support (Condor remote I/O), requires libchirp_client
        "cling              = ON ",  # ON  - Enable new CLING C++ interpreter
        # "cocoa              = *  ",  # *   - Use native Cocoa/Quartz graphics backend (MacOS X only)
        # "davix              = *  ",  # *   - DavIx library for HTTP/WEBDAV access
        "dcache             = OFF",  # ON  - dCache support, requires libdcap from DESY
        "exceptions         = ON ",  # ON  - Turn on compiler exception handling capability
        "explicit link      = *  ",  # *   - Explicitly link with all dependent libraries
        "fail-on-missing    = ON ",  # OFF - Fail the configure step if a required external package is missing
        "fftw3              = ON ",  # ON  - Fast Fourier Transform support, requires libfftw3
        "fitsio             = OFF",  # ON  - Read images and data from FITS files, requires cfitsio

        "gdml               = ON ",  # *   - GDML writer and reader
        "geocad             = OFF",  # OFF - ROOT-CAD Interface
        "genvector          = ON ",  # ON  - Build the new libGenVector library
        "gfal               = OFF",  # ON  - GFAL support, requires libgfal
        "glite              = OFF",  # ON  - gLite support, requires libglite-api-wrapper
        "globus             = OFF",  # OFF - Globus authentication support, requires Globus toolkit
        "gminimal           = OFF",  # OFF - Do not automatically search for support libraries, but include X11
        "gnuinstall         = OFF",  # OFF - Perform installation following the GNU guidelines
        "gsl_shared         = OFF",  # OFF - Enable linking against shared libraries for GSL (default no)
        "gviz               = OFF",  # ON  - Graphs visualization support, requires graphviz
        "hdfs               = OFF",  # ON  - HDFS support; requires libhdfs from HDFS >               =  0.19.1
        "http               = ON ",  # *   - HTTP Server support
        "imt                = ON ",  # ON  - Enable ROOT Multithreading Capabilities (default ON from version 6.10)
        "jemalloc           = OFF",  # OFF - Using the jemalloc allocator
        "krb5               = OFF",  # ON  - Kerberos5 support, requires Kerberos libs
        "ldap               = OFF",  # ON  - LDAP support, requires (Open)LDAP libs
        "mathmore           = ON ",  # ON  - Build the new libMathMore
        "memstat            = OFF",  # *   - A memory statistics utility, helps to detect memory leaks
        "minimal            = ON ",  # OFF - Do not automatically search for support libraries
        "minuit2            = ON ",  # *   - Build the new libMinuit2 minimizer library
        "monalisa           = OFF",  # ON  - Monalisa monitoring support, requires libapmoncpp
        "mt                 = OFF",  # OFF - Multi-threading support (deprecated and unused since ROOT v6.12)
        "mysql              = OFF",  # ON  - MySQL support, requires libmysqlclient
        "odbc               = OFF",  # ON  - ODBC support, requires libiodbc or libodbc
        "opengl             = OFF",  # ON  - OpenGL support, requires libGL and libGLU
        "oracle             = OFF",  # ON  - Oracle support, requires libocci
        "pgsql              = OFF",  # ON  - PostgreSQL support, requires libpq
        "pythia6            = OFF",  # ON  - Pythia6 EG support, requires libPythia6
        "pythia6_nolink     = OFF",  # OFF - Delayed linking of Pythia6 library
        "pythia8            = OFF",  # ON  - Pythia8 EG support, requires libPythia8
        "python             = ON ",  # ON  - Python ROOT bindings, requires python > = 2.2
        # "qt                    ",  # NA  - Qt graphics backend, requires libqt > = 4.8
        # "qtgsi              = *",  # *   - GSI's Qt integration, requires libqt > = 4.8
        "roofit             = ON ",  # *   - Build the libRooFit advanced fitting package
        "root7              = OFF",  # OFF - ROOT 7 support (read more)
        "roottest           = OFF",  # OFF - Include roottest in the test suit, if roottest exists in root
        "ruby               = OFF",  # OFF - Ruby ROOT bindings, requires ruby >= 1.8
        "r                  = OFF",  # OFF - R ROOT bindings, requires R, Rcpp and RInside
        "rfio               = OFF",  # ON  - RFIO support, requires libshift from CASTOR >= 1.5.2
        "rpath              = OFF",  # OFF - Set run-time library load path on executables and shared libraries
        "sapdb              = OFF",  # ON  - MaxDB/SapDB support, requires libsqlod and libsqlrte
        "shadowpw           = OFF",  # ON  - Shadow password support
        "shared             = OFF",  # ON  - Use shared 3rd party libraries if possible
        "soversion          = OFF",  # OFF - Set version number in sonames (recommended)
        "sqlite             = OFF",  # ON  - SQLite support, requires libsqlite3
        "srp                = OFF",  # ON  - SRP support, requires SRP source tree
        "ssl                = OFF",  # ON  - SSL encryption support, requires openssl
        "tbb                = OFF",  # OFF - TBB multi-threading support, requires TBB
        "table              = OFF",  # *   - Build libTable contrib library
        "tcmalloc           = OFF",  # OFF - Using the tcmalloc allocator
        "testing            = OFF",  # OFF - Enable test suit of ROOT with CTest
        "thread             = ON ",  # ON  - Using thread library (cannot be disabled)
        "tmva               = ON ",  # ON  - Build TMVA multi variate analysis library
        "unuran             = OFF",  # *   - UNURAN - package for generating non-uniform random numbers
        # "vc               = *  ",  # *   - Vc adds a few new types for portable and intuitive SIMD programming
        "vdt                = ON ",  # ON  - VDT adds a set of fast and vectorisable mathematical functions
        "winrtdebug         = OFF",  # OFF - Link against the Windows debug runtime library
        "xft                = OFF",  # ON  - Xft support (X11 antialiased fonts)
        "xml                = ON ",  # ON  - XML parser interface
        "xrootd             = OFF",  # ON  - Build xrootd file server and its client (if supported)
        "x11                = OFF",  # *   - X11 support

        "runtime_cxxmodules = OFF",  # OFF - Enable runtime c++ modules
    ]


def root_find():
    """Looks for CERN ROOT package
    :return [str] - empty list if not found or a list with 1 element - ROOTSYS path

    The only way to find ROOT is by checking for ROOTSYS package,
    The function family xxx_find() return list in general
    so this function returns ether empty list or a list with 1 element - root path
    """

    # The only way to find a CERN ROOT is by
    result = []

    # Check ROOTSYS environment variable
    if ROOTSYS not in os.environ:
        print("<red>ROOTSYS</red> not found in the environment")
        return result

    # Now check ROOTSYS exists in the system
    root_sys_path = os.environ[ROOTSYS]
    if not os.path.isdir(root_sys_path):
        print("WARNING", " ROOTSYS points to nonexistent directory of a file")
        return result

    # Looks like root exists, return the path
    return [root_sys_path]



# This module contains classes, which hold information about how to install packets
import os


class Recipe(object):
    """
    This class stores information which is pretty similar for each package (like directory structure)

    The important here is that while we try to stick to this naming, each package can override them if needed

    Usually this class is used like:

    # packets/my_packet.py file:

    MyPacketInstallationInstruction(PacketInstallationInstruction):
        # ... define your packet installation instruction by
        # defining:

        def set_app_path(...)  # - function that setups variables when the installation path is known

        # and logic with
        def step_install(...)  # - installs application
        def step_download(...) # - downloads from network
        def step...

    # usage of MyPacketInstallationInstruction is like:

    """

    def __init__(self, name):
        #
        # Short lowercase name of a packet
        self.name = name

        #
        # Next variables are set by ancestors
        self.config = {
            "app_name": self.name,
            "app_path"      : "",   # installation path of this packet, all other pathes relative to this
            "download_path" : "",   # where we download the source or clone git
            "source_path"   : "",   # The directory with source files for current version
            "build_path"    : "",   # The directory for cmake/scons build
            "install_path"  : "",   # The directory, where binary is installed
            "required_deps" : [],   # Packets which are required for this to run
            "optional_deps" : [],   # Optional packets
        }

    def setup(self, db):
        """This function is used to format and fill variables, when app_path is known download command"""
        # ... (!) inherited classes should implement its logic here
        raise NotImplementedError()

    def use_common_dirs_scheme(self):
        """Function sets common directory scheme. It is the same for many packets:
        """

        # where we download the source or clone git
        self.config["download_path"] = "{app_path}/src".format(app_path=self.app_path)

        # The directory with source files for current version
        self.config["source_path"] = "{app_path}/src/{branch}".format(**self.config)

        # The directory for cmake build
        self.config["build_path"] = "{app_path}/build/{branch}".format(**self.config)

        # The directory, where binary is installed
        self.config["install_path"] = "{app_path}/{app_name}-{branch}".format(**self.config)

    def step_install(self):
        """Installs application"""
        raise NotImplementedError()

    @property
    def app_path(self):
        return self.config["app_path"]

    @property
    def download_path(self):
        return self.config["download_path"]

    @property
    def source_path(self):
        return self.config["source_path"]

    @property
    def build_path(self):
        return self.config["build_path"]

    @property
    def install_path(self):
        return self.config["install_path"]

    @property
    def required_deps(self):
        return self.config["required_deps"]

    @required_deps.setter
    def required_deps(self, value):
        self.config["required_deps"] = value

    @property
    def optional_deps(self):
        return self.config["optional_deps"]

    @optional_deps.setter
    def optional_deps(self, value):
        self.config["optional_deps"] = value

    def source_dir_is_not_empty(self):
        return os.path.exists(self.source_path) \
               and os.path.isdir(self.source_path) \
               and os.listdir(self.source_path)



"""
Packet state database stores the known information about packets

Like: name, installation path, installation date,
"""

import json
import os
import io

from ejpm.engine.py23 import to_unicode


INSTALL_PATH = 'install_path'
SOURCE_PATH = 'source_path'
BUILD_PATH = 'build_path'
BUILT_WITH_CONFIG = 'built_with_config'
IS_OWNED = 'is_owned'
IS_ACTIVE = 'is_active'
IS_CDB = 'is_central_db'    # is from central database


# noinspection PyTypeChecker,PyUnresolvedReferences
class PacketStateDatabase(object):
    """Class to persist installation knowledge """

    def __init__(self):
        self.file_path = ""

        # This field is set during runtime by packet manager knowledge
        self.known_packet_names = []

        self.data = {
            "file_version": 3,  # This data structure version, each time will increase by 1
            "packets": {},      # create_packet_minimal_data() is used to create data inside this field
            "top_dir": "",      # The directory where everything is installed
            "packet_stack_name": "default",
            "global_build_config": {
                "build_threads": 4
            }
        }

        self.verbose = False
        self.is_loaded = False

    @staticmethod
    def create_packet_minimal_data():
        return {
            'build_config': {},
            'installs': []
        }
    
    def exists(self):
        """Returns True if db file exists

        self.file_path is used, it doesn't check is the file really exists or if it is a really our DB
        """
        return os.path.isfile(self.file_path)

    def save(self):
        """Saves self.data to a file self.file_path"""

        if not self.file_path:
            raise AssertionError()

        # Write JSON file
        with io.open(self.file_path, 'w', encoding='utf8') as outfile:
            str_ = json.dumps(self.data,
                              indent=4, sort_keys=True,
                              separators=(',', ': '), ensure_ascii=False)
            outfile.write(to_unicode(str_))

        self.is_loaded = True

    def load(self):
        """Loads self.data from a file with self.file_path"""

        # Read JSON file
        with open(self.file_path) as data_file:
            self.data = json.load(data_file)
            self.update_schema(self.data)

        self.is_loaded = True

    @property
    def packet_names(self):
        """Return packet names of known packets"""
        return self.data['packets'].keys()

    def get_installs(self, packet_name):
        """Get all installation information (like location) for packet with this name """

        # If we have no record in the packet installations, but we know the name is OK
        if packet_name not in self.data['packets'].keys() and packet_name in self.known_packet_names:
            self.data['packets'][packet_name] = PacketStateDatabase.create_packet_minimal_data()

        # Return whatever we can (or it will raise KeyException)
        return self.data['packets'][packet_name]['installs']

    def get_config(self, packet_name):
        """Get all installation information (like location) for packet with this name """

        # If we have no record in the packet installations, but we know the name is OK
        if packet_name not in self.data['packets'].keys() and packet_name in self.known_packet_names:
            self.data['packets'][packet_name] = PacketStateDatabase.create_packet_minimal_data()

        # Return whatever we can (or it will raise KeyException)
        if 'build_config' not in self.data['packets'][packet_name]:
            # noinspection PyUnresolvedReferences
            self.data['packets'][packet_name]['build_config'] = {}
        return self.data['packets'][packet_name]['build_config']

    def get_global_config(self):
        if 'global_build_config' not in self.data:
            self.data['global_build_config'] = {'build_threads': 4}

        return self.data['global_build_config']

    def get_active_install(self, packet_name):
        """Get installation marked as 'is_active' from all installations of packet with name 'packet_name'"""

        installs = self.get_installs(packet_name)

        for install in installs:
            assert isinstance(install, dict)
            if install[IS_ACTIVE]:
                return install

        return None

    def get_active_installs(self):
        """Returns {name:{install_data}} of active installs"""
        return {name: self.get_active_install(name) for name in self.packet_names}

    def get_install(self, packet_name, install_path):
        """
        Returns installation information for a given packet and a given path

        :param packet_name: Name of the packet like root, jana, etc
        :type packet_name: str
        :param install_path: Path of the packet installation
        :type install_path: str

        """
        installs = self.get_installs(packet_name)
        install_path = os.path.normpath(install_path)

        # Search for existing installation with this installation path
        for install in installs:
            # We compare it just by == as all saved installs have gone through os.path.normpath
            assert isinstance(install, dict)
            if install[INSTALL_PATH] == to_unicode(install_path):
                return install
        return None

    def update_install(self, packet_name, install_path, updating_data):
        """
        :param updating_data: A dict with data to update
        :type updating_data: dict
        :param packet_name: Name of the packet. Like root, genfit, rave, etc
        :param install_path: Path of the installation
        :return:
        """

        # normalize the path
        install_path = os.path.normpath(install_path)

        # Search for existing installation with this installation path
        existing_install = self.get_install(packet_name, install_path)

        # If we didn't find an install, lets add a new one
        if existing_install is None:
            existing_install = {}

            installs = self.get_installs(packet_name)
            installs.append(existing_install)
            existing_install[IS_ACTIVE] = True      # It is the first installation. Should be active
            existing_install[IS_OWNED] = False      # Nothing known about it, so guess we are not

        # if updating_data has IS_ACTIVE flag set to True, We have to uncheck all other our packets
        if updating_data.get(IS_ACTIVE, False):
            for install in self.get_installs(packet_name):
                # We compare it just by == as all saved installs have gone through os.path.normpath
                assert isinstance(install, dict)
                install[IS_ACTIVE] = False

        # set selected and ownership
        existing_install[INSTALL_PATH] = install_path
        existing_install.update(updating_data)
        return existing_install

    def remove_install(self, packet_name, install_path):
        """
        :param packet_name: Name of the packet. Like root, genfit, rave, etc
        :param install_path: Path of the installation
        :return:
        """

        installs = self.get_installs(packet_name)
        install_path = os.path.normpath(install_path)

        # Search for existing installation with this installation path
        for install in installs:
            # We compare it just by == as all saved installs have gone through os.path.normpath
            assert isinstance(install, dict)
            if install[INSTALL_PATH] == to_unicode(install_path):
                # noinspection PyUnresolvedReferences
                installs.remove(install)

    @property
    def top_dir(self):
        return self.data["top_dir"]

    @top_dir.setter
    def top_dir(self, path):
        self.data["top_dir"] = path

    def update_schema(self, data):
        """Do migration between different DB schema versions"""
        file_version = data['file_version']

        # version 1 to 2 migration
        if file_version == 1:
            # Remove some fields:
            # (Why pop https://stackoverflow.com/questions/11277432/how-to-remove-a-key-from-a-python-dictionary)

            # installed group (we use packet/installs instead)
            data.pop('installed', None)

            # 'required' (we use dependency chain instead)
            for name, packet in data['packets'].items():
                if 'required' in packet:
                    packet.pop('required', None)

            # finally change the version
            data['file_version'] = 2

        # version 2 to 3 migration
        if file_version == 2:

            data['global_build_config'] = {'build_threads': 4}

            for packet in data['packets'].values():
                packet['build_config'] = {}

            # finally change the version
            data['file_version'] = 3

        # version 3 to 4 migration
        if file_version == 3:
            if 'cxx_standard' not in data['global_build_config']:
                data['global_build_config']['cxx_standard'] = '11'

            data['file_version'] = 4


def merge_db(orig_db, import_db):
    assert isinstance(orig_db, PacketStateDatabase)
    assert isinstance(import_db, PacketStateDatabase)

    # Select names that are known for orig db
    import_names = [name for name in import_db.packet_names if name in orig_db.known_packet_names]

    for name in import_names:
        for import_install in import_db.get_installs(name):
            # Check if we have some installations of this packet
            no_prior_installs = orig_db.get_active_install(name) is None

            import_install[IS_ACTIVE] = True

            # Add imported record
            copied_install = orig_db.update_install(name, import_install[INSTALL_PATH], import_install)
            copied_install[IS_OWNED] = False

            # Set install active if we don't have an active install for this packet
#            copied_install[IS_ACTIVE] = no_prior_installs and import_install[IS_ACTIVE]
            # >oO debug pprint(copied_install)




import io
import os

import click

from ejpm.engine.api import pass_ejpm_context, EjpmApi, print_packets_info
from ejpm.engine.output import markup_print as mprint


# @click.group(invoke_without_command=True)
@click.command()
@click.argument('import_file', nargs=1)
@pass_ejpm_context
@click.pass_context
def mergedb(ctx, ectx, import_file):
    """Merges packet installation data into existing DB

    This command might be useful if one has DB installation and users
    would like to use centrally installed packages overriding some of them
    with new packages
    """

    assert isinstance(ectx, EjpmApi)

    # We need DB ready for this cli command
    ectx.ensure_db_exists()

    if not os.path.isfile(import_file):
        exit("Error! file '{}' does not exists or is not a file".format(import_file))


    # update_install will add or update the packet install. We set it active as it make sense...

    ectx.merge_external_db(import_file)
    ectx.db.save()

    # Update environment scripts
    mprint("Updating environment script files...\n")
    ectx.save_default_bash_environ()
    ectx.save_default_csh_environ()

    mprint("DB after the update:\n")

    print_packets_info(ectx.db)




"""
This file provides information of how to build and configure Delphes packet:
https://github.com/delphes/delphes.git

"""

import os

from ejpm.engine.env_gen import Prepend, Set, Append
from ejpm.engine.git_cmake_recipe import GitCmakeRecipe


class DelphesRecipe(GitCmakeRecipe):
    """Provides data for building and installing Delphes fast simulation framework"""

    def __init__(self):
        super(DelphesRecipe, self).__init__('delphes')
        self.config['branch'] = '3.4.2'
        self.config['repo_address'] = 'https://github.com/delphes/delphes.git'

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        path = data['install_path']

        yield Set("DELPHES_HOME", path)
        yield Prepend('CMAKE_PREFIX_PATH', os.path.join(path, 'share/eigen3/cmake/'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "",
            'centos': "",
            'centos8': ""
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



"""
This file provides information of how to build and configure Rave vertex finding packet:
https://github.com/WolfgangWaltenberger/rave

"""

import os

from ejpm.engine.env_gen import Set, Prepend
from ejpm.engine.recipe import Recipe
from ejpm.engine.commands import run, env, workdir


class RaveRecipe(Recipe):
    """Provides data for building and installing Rave vertex reconstruction package

    (PacketInstallationInstruction is located in recipe.py)
    """

    def __init__(self):
        # Call parent constructor to fill version, app_path ... and others
        # (!) it is called AFTER we override self.version
        super(RaveRecipe, self).__init__('rave')

        self.clone_command = ""
        self.bootstrap_command = ""         # This command is called after clone command
        self.build_command = ""
        self.config['branch'] = 'master'
        self.config['repo_address'] = 'https://github.com/WolfgangWaltenberger/rave.git'

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        #
        # use_common_dirs_scheme() sets standard package variables:
        # source_path  = {app_path}/src/{branch}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{branch}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{branch}         # Where the binary installation is
        self.use_common_dirs_scheme()

        # ENV RAVEPATH $INSTALL_DIR_RAVE

        # Git download link. Clone with shallow copy
        self.clone_command = "git clone --depth 1 -b {branch} {repo_address} {source_path}".format(**self.config)

        # cmake command:
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        # (!) {{clhep_lib_dir}} and {{clhep_include_dir}} is in 2 braces. So it becomes {clhep_include_dir}
        #     on step_build the right environment will be set and we will use .format(..) again to fill them
        #
        #     ugly? YEA!!! say hello to RAVE
        # (!)
        self.build_command = './configure --disable-java --prefix=$RAVEPATH ' \
                             ' LDFLAGS="-L{{clhep_lib_dir}}" ' \
                             ' CXXFLAGS="-std=c++11 -I{{clhep_include_dir}}" ' \
                             '&&   make -j{build_threads} install' \
                             "&& for f in $(ls $RAVEPATH/include/rave/*.h); do sed -i 's/RaveDllExport//g' $f; done" \
            .format(**self.config)

    def step_install(self):
        self.step_clone()
        self.step_build()

    def step_clone(self):
        """Clones RAVE from github mirror"""

        # Check the directory exists and not empty
        if self.source_dir_is_not_empty():
            # The directory exists and is not empty. Nothing to do
            return
        else:
            # Create the directory
            run('mkdir -p {}'.format(self.source_path))

        run(self.clone_command)         # Execute git clone command
        workdir(self.source_path)       # Go to our build directory
        run('./bootstrap')              # This command required to run by rave once...

    def step_build(self):
        # Create build directory

        env('RAVEPATH', self.install_path)

        # Rave uses ./configure so we building it in {source_path}
        # go to our build directory
        workdir(self.source_path)

        # environment variables CLHEP_LIB_DIR and CLHEP_INCLUDE_DIR should be set at this stage to compile it
        # otherwise we fall back to default debian install location (RHEL7 doesn't have clep in its repo)
        clhep_lib_dir = os.environ.get('CLHEP_LIB_DIR', '/usr/lib')
        clhep_include_dir = os.environ.get('CLHEP_INCLUDE_DIR', '/usr/include')

        build_command_with_clhep = self.build_command.format(
            clhep_lib_dir=clhep_lib_dir,
            clhep_include_dir=clhep_include_dir)

        # run cmake && make && install
        run(build_command_with_clhep)

    @staticmethod
    def gen_env(data):
        install_path = data['install_path']

        yield Set('RAVEPATH', install_path)
        yield Prepend('CMAKE_PREFIX_PATH', os.path.join(install_path, 'share', 'rave'))
        yield Prepend('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "autoconf shtool automake libtool",
            'centos': "autoconf shtool automake libtool"
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



"""
This file provides information of how to build and configure Fast Jet:
http://fastjet.fr

"""

import os

from ejpm.engine.env_gen import Set, Prepend
from ejpm.engine.recipe import Recipe
from ejpm.engine.commands import run, env, workdir


class FastJetRecipe(Recipe):
    """Provides data for building and installing Rave vertex reconstruction package

    (PacketInstallationInstruction is located in recipe.py)
    """

    def __init__(self):
        # Call parent constructor to fill version, app_path ... and others
        # (!) it is called AFTER we override self.version
        super(FastJetRecipe, self).__init__('fastjet')

        self.clone_command = ""
        self.unpack_command = ""         # This command is to untar downloaded array
        self.build_command = ""
        self.config['branch'] = '3.3.3'
        self.config['archive_name'] = 'fastjet-{branch}'
        self.config['repo_address'] = 'http://fastjet.fr/repo/{archive_name}.tar.gz'

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        #
        # use_common_dirs_scheme() sets standard package variables:
        # source_path  = {app_path}/src/{branch}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{branch}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{branch}         # Where the binary installation is
        self.use_common_dirs_scheme()

        # ENV RAVEPATH $INSTALL_DIR_RAVE

        # Git download link. Clone with shallow copy
        self.config['archive_name'] = self.config['archive_name'].format(**self.config)
        self.config['repo_address'] = self.config['repo_address'].format(**self.config)
        self.clone_command = "curl -O {repo_address}".format(**self.config)
        self.unpack_command = 'tar zxvf {archive_name}.tar.gz&& mv {archive_name}/* {source_path}'.format(**self.config)
        self.build_command = './configure --prefix={install_path} &&' \
                             ' make -j{build_threads} &&' \
                             ' make check && make install' \
            .format(**self.config)

    def step_install(self):
        self.step_clone()
        self.step_build()

    def step_clone(self):
        """Clones RAVE from github mirror"""

        # Check the directory exists and not empty
        if self.source_dir_is_not_empty():
            # The directory exists and is not empty. Nothing to do
            return
        else:
            # Create the directory
            run('mkdir -p {}'.format(self.source_path))

        workdir(self.source_path)  # Go to our build directory

        run(self.clone_command)         # Execute git clone command
        run(self.unpack_command)         # Execute git clone command

    def step_build(self):
        # go to our build directory
        workdir(self.source_path)

        # run cmake && make && install
        run(self.build_command)

    @staticmethod
    def gen_env(data):
        install_path = data['install_path']
        yield Prepend('PATH', os.path.join(install_path, 'bin'))
        yield Prepend('LD_LIBRARY_PATH', os.path.join(install_path, 'lib'))

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "",
            'centos': ""
        },
        'optional': {
            'ubuntu': "",
            'centos': ""
        },
    }



import os

class EnvironmentManipulation(object):
    def __init__(self, name, value):
        self.name = name
        self.value = value

    def gen_bash(self):
        """Generates bash piece of code"""
        raise NotImplementedError()

    def gen_csh(self):
        """Generates csh piece code"""
        raise NotImplementedError()

    def update_python_env(self):
        """Sets environment internally for python"""
        raise NotImplementedError()

    @staticmethod
    def is_in_path_env(path):
        """Sets environment internally for python"""
        path_tokens = os.environ.get('PATH', '').split(os.pathsep)
        return path in path_tokens


class Append(EnvironmentManipulation):
    """Appends environment variables like PATH or LD_LIBRARY_PATH"""

    def __init__(self, name, value):
        super(Append, self).__init__(name, value)

    def gen_bash(self):
        """Generates bash piece of code"""

        # Here is the right way to APPEND in bash:
        # PATH = "${PATH:+${PATH}:}$HOME/bin"
        # https://unix.stackexchange.com/a/415028/109031
        ret_str = "export {name}=${{{name}:+${{{name}}}:}}{value}\n"

        #for appending(instead of PATH="$PATH:$HOME/bin") and
        # ret_str = (
        #     '\n'
        #     '# Make sure {name} is set\n'
        #     'if [ -z "${name}" ]; then\n'
        #     '    export {name}="{value}"\n'
        #     'else\n'
        #     '    export {name}=${name}:"{value}"\n'
        #     'fi')

        return ret_str.format(name=self.name, value=self.value)

    def gen_csh(self):
        """Generates csh piece code"""

        ret_str = (
            '\n'
            '# Make sure {name} is set\n'
            'if ( ! $?{name} ) then\n'
            '    setenv {name} "{value}"\n'
            'else\n'
            '    setenv {name} ${{{name}}}:"{value}"\n'
            'endif')

        return ret_str.format(name=self.name, value=self.value)

    def update_python_env(self):
        """Sets environment internally for python"""
        print("   update_env:  append ${} by '{}'".format(self.name, self.value))
        if self.name in os.environ.keys():
            os.environ[self.name] += os.pathsep + self.value
        else:
            os.environ[self.name] = self.value


class Prepend(EnvironmentManipulation):
    """Prepends environment variables like PATH or LD_LIBRARY_PATH"""

    def __init__(self, name, value):
        super(Prepend, self).__init__(name, value)

    def gen_bash(self):
        """Generates bash piece of code"""

        # Here is the right way to prepend in bash:
        # #PATH = "$HOME/bin${PATH:+:${PATH}}"
        # https://unix.stackexchange.com/a/415028/109031
        ret_str = "export {name}={value}${{{name}:+:${{{name}}}}}\n"

        # ret_str = (
        #     '\n'
        #     '# Make sure {name} is set\n'
        #     'if [ -z "${name}" ]; then\n'
        #     '    export {name}="{value}"\n'
        #     'else\n'
        #     '    export {name}="{value}":${name}\n'
        #     'fi')

        return ret_str.format(name=self.name, value=self.value)

    def gen_csh(self):
        """Generates csh piece code"""

        ret_str = (
            '\n'
            '# Make sure {name} is set\n'
            'if ( ! $?{name} ) then\n'
            '    setenv {name} "{value}"\n'
            'else\n'
            '    setenv {name} "{value}":${{{name}}}\n'
            'endif')

        return ret_str.format(name=self.name, value=self.value)

    def update_python_env(self):
        """Sets environment internally for python"""

        print("   update_env: prepend ${} by '{}'".format(self.name, self.value))
        if self.name in os.environ.keys():
            os.environ[self.name] = self.value + os.pathsep + os.environ[self.name]
        else:
            os.environ[self.name] = self.value


class Set(EnvironmentManipulation):
    """Sets environment variables like PATH or LD_LIBRARY_PATH"""

    def __init__(self, name, value):
        super(Set, self).__init__(name, value)

    def gen_bash(self):
        """Generates bash piece of code"""
        return 'export {name}="{value}"'.format(name=self.name, value=self.value)

    def gen_csh(self):
        """Generates csh piece code"""
        return 'setenv {name} "{value}"'.format(name=self.name, value=self.value)

    def update_python_env(self):
        """Sets environment internally for python"""

        print("   update_env:     set ${} = '{}'".format(self.name, self.value))
        os.environ[self.name] = self.value


class RawText(EnvironmentManipulation):
    def __init__(self, sh_text, csh_text, python_env):
        """ Function allows to

        :param sh_text: Text for Bash generated script
        :param csh_text: Text for CSH generated script
        :param python_env: The function that manipulates python environment right
        """

        super(RawText, self).__init__('', '')
        self.sh_text = sh_text
        self.csh_text = csh_text
        self.python_env = python_env

    def gen_bash(self):
        """Generates bash piece of code"""
        return self.sh_text

    def gen_csh(self):
        """Generates csh piece code"""
        return self.csh_text

    def update_python_env(self):
        """Sets environment internally for python"""
        print("   update_env: RawText step will update env")
        self.python_env()


class Comment(EnvironmentManipulation):
    def __init__(self, text):
        """
        Adds comment to generated script

        :param text: The comment text
        """
        super(Comment, self).__init__('', '')
        self.text = text

    def gen_bash(self):
        """Generates bash piece of code"""
        return self.text

    def gen_csh(self):
        """Generates csh piece code"""
        return self.text

    def update_python_env(self):
        """Sets environment internally for python"""
        pass    # Just nothing to do!



"""
This file provides information of how to build and configure EIC Jana (ejana) framework:
https://gitlab.com/eic/ejana

"""

import os

from ejpm.engine.env_gen import Set, Prepend
from ejpm.engine.recipe import Recipe
from ejpm.engine.commands import run, env, workdir


class EjanaRecipe(Recipe):
    """Provides data for building and installing JANA framework

    PackageInstallationContext is located in recipe.py and contains the next standard package variables:

    version      = 'v{}-{:02}-{:02}'                 # Stringified version. Used to create directories and so on
    glb_app_path = Context.work_dir                  # The directory where all other packets are installed
    source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
    build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
    install_path = {app_path}/root-{version}         # Where the binary installation is
    """


    def __init__(self):
        """
        """
        super(EjanaRecipe, self).__init__('ejana')
        self.clone_command = ""
        self.build_command = ""
        self.required_deps = ['root', 'hepmc3', 'eic-smear', 'jana']
        self.config['branch'] = 'v1.2.3'
        self.config['install_mode'] = 'dev'

    def _setup_dev(self):
        # The directory with source files for current version
        self.config['source_path'] = "{app_path}/ejana-dev".format(**self.config)
        self.config['build_path'] = "{app_path}/ejana-dev/cmake-build-debug".format(**self.config)  # build in dev directory
        self.config['install_path'] = "{app_path}/ejana-dev/compiled".format(**self.config)

        #
        # ejana download link
        self.clone_command = "git clone -b {branch} https://gitlab.com/eic/ejana.git {source_path}"\
                             .format(**self.config)

        # cmake command:
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev -DCMAKE_INSTALL_PREFIX={install_path} -DCMAKE_CXX_STANDARD={cxx_standard} {source_path}" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
                         .format(**self.config)

    def setup(self, db):
        """Sets all variables like source dirs, build dirs, etc"""

        # (!) at this point we alwais use dev environment
        if self.config['install_mode'] == 'dev':
            return self._setup_dev()

        #
        # use_common_dirs_scheme sets standard package variables:
        # version      = 'v{}-{:02}-{:02}'                 # Stringified version. Used to create directories and so on
        # source_path  = {app_path}/src/{version}          # Where the sources for the current version are located
        # build_path   = {app_path}/build/{version}        # Where sources are built. Kind of temporary dir
        # install_path = {app_path}/root-{version}         # Where the binary installation is
        self.use_common_dirs_scheme()

        #
        # Git download link. Clone with shallow copy
        self.clone_command = "git clone --depth 1 -b {branch} https://gitlab.com/eic/escalate/ejana.git {source_path}"\
            .format(**self.config)

        # cmake command:
        # the  -Wno-dev  flag is to ignore the project developers cmake warnings for policy CMP0075
        self.build_cmd = "cmake -Wno-dev -DCMAKE_INSTALL_PREFIX={install_path} {source_path}" \
                         "&& cmake --build . -- -j {build_threads}" \
                         "&& cmake --build . --target install" \
                         .format(**self.config)  # make global options like '-j8'. Skip now

    def step_install(self):
        self.step_clone()
        self.step_build()

    def step_clone(self):
        """Clones JANA from github mirror"""

        # Check the directory exists and not empty
        if self.source_dir_is_not_empty():
            return  # The directory exists and is not empty. Nothing to do

        run('mkdir -p {source_path}'.format(**self.config))   # Create the directory
        run(self.clone_command)                               # Execute git clone command

    def step_build(self):
        """Builds JANA from the ground"""

        # Create build directory
        run('mkdir -p {}'.format(self.build_path))

        # go to our build directory
        workdir(self.build_path)

        # run scons && scons install
        run(self.build_cmd)

    def step_reinstall(self):
        """Delete everything and start over"""

        # clear sources directories if needed
        run('rm -rf {}'.format(self.app_path))

        # Now run build root
        self.step_install()

    @staticmethod
    def gen_env(data):
        """Generates environments to be set"""
        install_path = data['install_path']
        yield Set('EJANA_HOME', install_path)
        yield Prepend('JANA_PLUGIN_PATH', os.path.join(install_path, 'plugins'))
        yield Prepend('PATH', os.path.join(install_path, 'bin'))

        lib_path = os.path.join(install_path, 'lib')  # on some platforms
        lib64_path = os.path.join(install_path, 'lib64')  # on some platforms

        if os.path.isdir(lib64_path):
            yield Prepend('LD_LIBRARY_PATH', lib64_path)
        else:
            yield Prepend('LD_LIBRARY_PATH', lib_path)

    #
    # OS dependencies are a map of software packets installed by os maintainers
    # The map should be in form:
    # os_dependencies = { 'required': {'ubuntu': "space separated packet names", 'centos': "..."},
    #                     'optional': {'ubuntu': "space separated packet names", 'centos': "..."}
    # The idea behind is to generate easy to use instructions: 'sudo apt-get install ... ... ... '
    os_dependencies = {
        'required': {
            'ubuntu': "",
            'centos': "",
            'centos8':""
        },
        'optional': {
            'ubuntu': "",
            'centos': "",
            'centos8':""
        },
    }

    # Flags that can me made in cmake
    cmake_deps_flag_names = {
        "root": "ROOT_DIR",             # Cern root installation
        "jana": 'JANA_DIR',             # JANA2 installation directory
        'genfit': 'GENFIT_DIR',         # Genfit2  installation directory
        'eic-smear': 'EIC_SMEAR_DIR',   # EIC-smear smearing packet
        'hepmc': 'HEPMC_DIR'
    }

